//+------------------------------------------------------------------+
//|                                               HistoryDumpCSV.mq4 |
//|                                       Copyright 2016, Mark Flint |
//|                            https://www.mql5.com/en/users/flima02 |
//+------------------------------------------------------------------+
#property copyright "Copyright 2016, Mark Flint"
#property link      "https://www.mql5.com/en/users/flima02"
#property version   "1.02"
#property strict
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#property description "This script creates a .CSV file of your order"
#property description "history so you can import it into Excel and analyse it."
#property description "It extracts date/time information in both"
#property description "MetaTrader format (_MT) and Excel format."
#property description "It supports aggregate analysis by calculating"
#property description "a 'Heat map' day and time slot."
#property description "The heat map default is 4 slots per hour i.e. 15 minutes."
#property description "The script will FTP the file to your specified"
#property description "FTP server, if you have defined one."
#property description "Records are sorted in Date/Time Closed order."
#property description "Times are server times. File is tab delimited."

//+------------------------------------------------------------------+
//| structure of history order entry                                 |
//+------------------------------------------------------------------+
struct Orders
  {
   int               SequenceNumber;
   double            Point; // not output
   string            AccCompany;
   string            AccServer;
   string            AccName;
   int               AccNumber;
   string            ClosePrice; //double
   string            ServerCloseTime;
   datetime          ServerCloseTime_MT;
   string            Comment;
   string            Commission; //double
   string            ServerExpiration;
   datetime          ServerExpiration_MT;
   string            Lots; // double
   int               MagicNumber;
   string            OpenPrice; // double
   string            ServerOpenTime;
   datetime          ServerOpenTime_MT;
   string            NetProfit; //double // with swaps and commissions
   string            Profit; //double // without swaps and commissions
   string            StopLoss;//double
   string            Swap;//double
   string            Symbol;
   string            TakeProfit;//double
   long              Ticket; // ID
   string            Type; // OP_BUY, OP_SELLSTOP etc.
   string            PriceMove;
   string            Points;
   string            Pips;//double
   int               DurationMins;
   int               HeatMapOpenDay;
   int               HeatMapOpenSlot;
   int               HeatMapOpenHour;
   int               HeatMapCloseSlot;
   int               HeatMapCloseDay;
   int               HeatMapCloseHour;
  };
//+------------------------------------------------------------------+
//| Sorting structure                                                |
//+------------------------------------------------------------------+
struct TicketClose
  {
   int               TicketID;
   datetime          ServerCloseTime;
  };

   // Array for sorting tickets into date/time closed sequence.
   TicketClose ClosedTickets[];
   TicketClose TempClosedTicket;


string AccountID=IntegerToString(AccountNumber());
string reportfilename=StringConcatenate("HistoryDumpCSV_",AccountID,".CSV");  // the FileOpen has the FILE_COMMON flag meaning it is saved to the Commong/Files folder.
uchar DELIM='\t';  // Delimiter is '\t' tab.  Other options are ';' semicolon and ',' comma

int filehandle;
bool NeedToOpenFile=true; // Set this to allow the process to open file and write the header.
int SequenceNum=0;

bool SendFileByFTP=true; // If not needed, set this to false.

string msg="";
MqlDateTime ServerMqlDateTimeGMT;
long ServerTimeGMTOffsetSeconds=0;
int SLOTSPERHOUR=4;  // This divides the trading time into 15 minute slots for aggregating data.  2=30 minutes, 3=20 minutes, 1=1hour slots.
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---
   msg=StringConcatenate("HistoryDumpCSV: Save history data to ",reportfilename);
   MessageBox(msg);
   int i,j;
   bool FoundOrder,SavedDataOK;
   int total=OrdersHistoryTotal();
   Orders ThisOrder;
   
   ArrayResize(ClosedTickets,total);
   
   //ServerTimeGMTOffsetSeconds=TimeGMT()-TimeCurrent(); // negative if behind GMT, positive if ahead of GMT.

                                                       // Populate sorting array
   for(i=0; i<=total; i++)
     {
      FoundOrder=OrderSelect(i,SELECT_BY_POS,MODE_HISTORY);
      if(!FoundOrder) break;
      ClosedTickets[i].TicketID=OrderTicket();
      ClosedTickets[i].ServerCloseTime=OrderCloseTime();
     };
//Bubble sort array
   for(i=0; i<total-1; i++)
     {
      for(j=i+1; j<total; j++)
        {
         if(ClosedTickets[j].ServerCloseTime<ClosedTickets[i].ServerCloseTime)
           {
            // swap the entries
            TempClosedTicket.ServerCloseTime= ClosedTickets[i].ServerCloseTime;
            TempClosedTicket.TicketID       = ClosedTickets[i].TicketID;

            ClosedTickets[i].ServerCloseTime = ClosedTickets[j].ServerCloseTime;
            ClosedTickets[i].TicketID        = ClosedTickets[j].TicketID;

            ClosedTickets[j].ServerCloseTime = TempClosedTicket.ServerCloseTime;
            ClosedTickets[j].TicketID        = TempClosedTicket.TicketID;
           }
        }

     }

   for(i=0; i<total; i++)
     {
      //FoundOrder=OrderSelect(i,SELECT_BY_POS,MODE_HISTORY);
      FoundOrder=OrderSelect(ClosedTickets[i].TicketID,SELECT_BY_TICKET,MODE_HISTORY);
      FillOrderData(ThisOrder);
      if(!FoundOrder || ThisOrder.Ticket<1) continue;
      SavedDataOK=SaveData(ThisOrder);
      if(!SavedDataOK) break; // error, exit.
     };
   if(!NeedToOpenFile) // file was openned, so now close it
     {
      FileFlush(filehandle); // make sure it is flushed to disk before we exit
      FileClose(filehandle);
      if(SendFileByFTP)
        {
         if(!SendFTP(reportfilename))
           {
            PlaySound("alert.wav");
            msg=StringConcatenate("HistoryDumpCSV: ERROR: FTP Failed to send file ",reportfilename);
            Print(msg); // log error in Journal
            MessageBox(msg); // interrupt user
            return;
           }
         else
           {
            msg="HistoryDumpCSV: File generated and successfully sent by FTP.";
            Print(msg); // log to Journal
            PlaySound("ok.wav"); // file generated and successfully sent by FTP.  OK!
           }
        }
      else
        {
         msg="HistoryDumpCSV: File generated successfully.  Not sent by FTP as per settings.";
         Print(msg); // log to Journal
         PlaySound("ok.wav"); // file generated successfully. FTP not wanted. OK!
        }
     }
   else
     {
      PlaySound("alert.wav");
      msg="HistoryDumpCSV: File open error / No file generated.";
      Print(msg); // log to Journal
      MessageBox(msg); // interrupt user
     }
   return;
  }
//+------------------------------------------------------------------+
//|  Fill Orders structure object                                    |
//+------------------------------------------------------------------+
void FillOrderData(Orders &ThisOrder)
  {
// Use immediately after OrderSelect, OrderSend
   ThisOrder.SequenceNumber=++SequenceNum;
   ThisOrder.Point=SymbolInfoDouble(OrderSymbol(),SYMBOL_POINT);
   ThisOrder.AccCompany=CleanString(AccountCompany());
   ThisOrder.AccServer=CleanString(AccountServer());
   ThisOrder.AccName=CleanString(AccountName());
   ThisOrder.AccNumber=AccountNumber();
   ThisOrder.ClosePrice=DoubleToString(OrderClosePrice(),Digits);
   ThisOrder.ServerCloseTime=CleanDateTime(OrderCloseTime());
   ThisOrder.ServerCloseTime_MT=OrderCloseTime(); // native MT
   ThisOrder.Comment=CleanString(OrderComment());
   ThisOrder.Commission=DoubleToString(OrderCommission(),2);
   ThisOrder.ServerExpiration=CleanDateTime(OrderExpiration());
   ThisOrder.ServerExpiration_MT=OrderExpiration(); // native MT
   ThisOrder.Lots=DoubleToString(OrderLots(),2);
   ThisOrder.MagicNumber=OrderMagicNumber();
   ThisOrder.OpenPrice=DoubleToString(OrderOpenPrice(),Digits);
   ThisOrder.ServerOpenTime=CleanDateTime(OrderOpenTime());
   ThisOrder.ServerOpenTime_MT=OrderOpenTime();
   ThisOrder.NetProfit=DoubleToString(OrderProfit()+OrderSwap()+OrderCommission(),3); // profit minus swaps and commissions
   ThisOrder.Profit=DoubleToString(OrderProfit(),3); // without swaps and commissions
   ThisOrder.StopLoss=DoubleToString(OrderStopLoss(),Digits);
   ThisOrder.Swap=DoubleToString(OrderSwap(),Digits);
   ThisOrder.Symbol=OrderSymbol();
   ThisOrder.TakeProfit=DoubleToString(OrderTakeProfit(),Digits);
   ThisOrder.Ticket=OrderTicket(); // ID
   ThisOrder.Type=OrderTypeName(OrderType()); // string name from int for OP_BUY, OP_SELLSTOP etc.
   ThisOrder.PriceMove=DoubleToString(NormalizeDouble(OrderClosePrice()-OrderOpenPrice(),Digits),Digits);
   ThisOrder.Points="0.000";
   if(ThisOrder.Type=="OP_BUY")  ThisOrder.Points=DoubleToString(NormalizeDouble(PriceMoveToPoints(OrderClosePrice()-OrderOpenPrice(),ThisOrder.Point),Digits),Digits);
   if(ThisOrder.Type=="OP_SELL") ThisOrder.Points=DoubleToString(NormalizeDouble(PriceMoveToPoints(OrderOpenPrice()-OrderClosePrice(),ThisOrder.Point),Digits),Digits);
   ThisOrder.Pips="0.000";
   if(ThisOrder.Type=="OP_BUY")  ThisOrder.Pips=DoubleToString(NormalizeDouble(PointsToPips(PriceMoveToPoints(OrderClosePrice()-OrderOpenPrice(),ThisOrder.Point)),Digits),Digits);
   if(ThisOrder.Type=="OP_SELL") ThisOrder.Pips=DoubleToString(NormalizeDouble(PointsToPips(PriceMoveToPoints(OrderOpenPrice()-OrderClosePrice(),ThisOrder.Point)),Digits),Digits);
   ThisOrder.DurationMins=int((double(OrderCloseTime())-double(OrderOpenTime()))/60);
   ThisOrder.HeatMapOpenDay=(SlotDayFromServerTime(OrderOpenTime()));
   ThisOrder.HeatMapOpenSlot=(SlotNumFromServerTime(OrderOpenTime()));
   ThisOrder.HeatMapOpenHour=(TimeHour(OrderOpenTime()));
   ThisOrder.HeatMapCloseDay=(SlotDayFromServerTime(OrderCloseTime()));
   ThisOrder.HeatMapCloseSlot=(SlotNumFromServerTime(OrderCloseTime()));
   ThisOrder.HeatMapCloseHour=(TimeHour(OrderCloseTime()));
   return;
  }
//+------------------------------------------------------------------+
bool SaveData(Orders &ThisOrder)
  {
//Save it to disk...
   string CleanComment;

   if(NeedToOpenFile)
     {
      ResetLastError();
      // filehandle=FileOpen(reportfilename,FILE_COMMON|FILE_WRITE|FILE_CSV); // overwrite previous file if present.  If written to the COMMON files directory, the SendFTP function will not work.
      filehandle=FileOpen(reportfilename,FILE_WRITE|FILE_CSV,DELIM); // overwrite previous file if present.  Send to local files directory.  SendFTP should work.
      if(filehandle==INVALID_HANDLE || filehandle<=0) return(false); // we can't open the file so exit the try.
     }
   if(filehandle!=INVALID_HANDLE)
     {
      if(NeedToOpenFile) // write a header line first.
        {
         FileWrite(filehandle,
                   "AccCompany",
                   "AccServer",
                   "AccName",
                   "AccNumber",
                   "SeqNum",
                   "ClosePrice",
                   "ServerCloseTime",
                   "ServerCloseTime_MT",
                   "Comment",
                   "Commission",
                   "ServerExpiration",
                   "ServerExpiration_MT",
                   "Lots",
                   "MagicNumber",
                   "OpenPrice",
                   "ServerOpenTime",
                   "ServerOpenTime_MT",
                   "NetProfit",
                   "Profit",
                   "StopLoss",
                   "Swap",
                   "Symbol",
                   "TakeProfit",
                   "Ticket",
                   "Type",
                   "PriceMove",
                   "Points",
                   "Pips",
                   "DurationMins",
                   "HeatMapOpenDay0to6",
                   "HeatMapOpenSlot",
                   "HeatMapOpenHour",
                   "HeatMapCloseDay0to6",
                   "HeatMapCloseSlot",
                   "HeatMapCloseHour" );
         NeedToOpenFile=false; // Dont need to open it or write header line again
        }

      FileWrite(filehandle,
                ThisOrder.AccCompany,
                ThisOrder.AccServer,
                ThisOrder.AccName,
                ThisOrder.AccNumber,
                ThisOrder.SequenceNumber,
                ThisOrder.ClosePrice,
                ThisOrder.ServerCloseTime,
                ThisOrder.ServerCloseTime_MT,
                ThisOrder.Comment,
                ThisOrder.Commission,
                ThisOrder.ServerExpiration,
                ThisOrder.ServerExpiration_MT,
                ThisOrder.Lots,
                ThisOrder.MagicNumber,
                ThisOrder.OpenPrice,
                ThisOrder.ServerOpenTime,
                ThisOrder.ServerOpenTime_MT,
                ThisOrder.NetProfit,
                ThisOrder.Profit,
                ThisOrder.StopLoss,
                ThisOrder.Swap,
                ThisOrder.Symbol,
                ThisOrder.TakeProfit,
                ThisOrder.Ticket,
                ThisOrder.Type,
                ThisOrder.PriceMove,
                ThisOrder.Points,
                ThisOrder.Pips,
                ThisOrder.DurationMins,
                ThisOrder.HeatMapOpenDay,
                ThisOrder.HeatMapOpenSlot,
                ThisOrder.HeatMapOpenHour,
                ThisOrder.HeatMapCloseDay,
                ThisOrder.HeatMapCloseSlot,
                ThisOrder.HeatMapCloseHour  );
      return(true);
     }
   else
     {
      MessageBox(StringConcatenate("Operation FileOpen failed for ",reportfilename,", error ",GetLastError()));
      return(false);
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Make string safe for .CSV files using comma, tab or semi-colon as seperator
//+------------------------------------------------------------------+
string CleanString(string s)
  {
   string temp=s;
   StringReplace(temp,";","|"); // replace any ';' with '|' to avoid the field delimiter in the .csv file
   StringReplace(temp,",","|"); // replace any ',' with '|' to avoid the field delimiter in the .csv file
   StringReplace(temp,"\t"," "); // replace any {tab} with {space} to avoid the field delimiter in the .csv file
   return(temp);
  }
//+------------------------------------------------------------------+
//| Text name of order type
//+------------------------------------------------------------------+
string OrderTypeName(int TradeOperation)
  {
   switch(TradeOperation)
     {
      case 0 : return("OP_BUY");
      case 1 : return("OP_SELL");
      case 2 : return("OP_BUYLIMIT");
      case 3 : return("OP_SELLLIMIT");
      case 4 : return("OP_BUYSTOP");
      case 5 : return("OP_SELLSTOP");
      default : return("other");
     }
   return("unknown?");
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string CleanDateTime(datetime d)
  {
   return(StringConcatenate(CleanDate(d)," ",CleanTime(d)));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string CleanDate(datetime d)
  {
   string temp;

   temp=StringConcatenate("00",IntegerToString(TimeYear(d)));
   string yyyy=StringSubstr(temp,(StringLen(temp)-4),4);

   temp=StringConcatenate("00",IntegerToString(TimeMonth(d)));
   string mm=StringSubstr(temp,(StringLen(temp)-2),2);

   temp=StringConcatenate("00",IntegerToString(TimeDay(d)));
   string dd=StringSubstr(temp,(StringLen(temp)-2),2);

   return(StringConcatenate(dd,"-",mm,"-",yyyy));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string CleanTime(datetime d)
  {
   string temp;

   temp=StringConcatenate("00",IntegerToString(TimeHour(d)));
   string hh=StringSubstr(temp,(StringLen(temp)-2),2);

   temp=StringConcatenate("00",IntegerToString(TimeMinute(d)));
   string mi=StringSubstr(temp,(StringLen(temp)-2),2);

   temp=StringConcatenate("00",IntegerToString(TimeSeconds(d)));
   string ss=StringSubstr(temp,(StringLen(temp)-2),2);

   return(StringConcatenate(hh,":",mi,":",ss));
  }
//+------------------------------------------------------------------+
//| Translate a price move (Buy-Sell) into points for the currency
//+------------------------------------------------------------------+
double PriceMoveToPoints(double pricemove,double point)
  {
   double p=point;
   if(p==0) p=SymbolInfoDouble(Symbol(),SYMBOL_POINT);

   if(Digits==3 || Digits==5) return((pricemove/p)*1);
   return((pricemove/p)*10);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double PointsToPips(double points)
  {
   if(Digits==3 || Digits==5) return(points/10);
   return(points);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double PipsToPoints(double pips)
  {
   if(Digits==3 || Digits==5) return(pips*10);
   return(pips);
  }
//+------------------------------------------------------------------+
//| Slot Day from GMT
//+------------------------------------------------------------------+
int SlotNumFromServerTime(datetime T=0)
  {
   if(T==0) T=TimeCurrent(); // last known server time
   int th,m;
   th=TimeHour(T+ServerTimeGMTOffsetSeconds); // hour number 0-23 on the server in GMT
   m=(int)MathFloor(TimeMinute(T+ServerTimeGMTOffsetSeconds)/(60/SLOTSPERHOUR));  // should yield 0-3 for 15 minute slots when SLOTSPERHOUR=4
   return((th*SLOTSPERHOUR)+m); // slot number of current GMT
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int SlotDayFromServerTime(datetime T=0)
  {
   if(T==0) T=TimeCurrent(); // last known server time
   return(TimeDayOfWeek(T+ServerTimeGMTOffsetSeconds)); // 0=Sunday, 1=Monday etc. 5=Friday, 6=Saturday
  }
//+------------------------------------------------------------------+
